#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from collections import defaultdict
import itertools
import networkx as nx

from .utils import maxes, first_alpha, are_all_equal


def add_element_attr(molecule):
    """
    Adds an element attribute to every node in `molecule`, based on that node's
    atomname attribute.

    Parameters
    ----------
    molecule: networkx.Graph
        The graph of which nodes should get an element attribute.

    Raises
    ------
    ValueError
        If no element could be guessed for a node.

    """
    for node_idx in molecule:
        node = molecule.nodes[node_idx]
        if 'element' not in node:
            try:
                element = first_alpha(node['atomname'])
            except KeyError:
                raise ValueError('Cannot guess the element of atom {}: '
                                 'the node has no atom name.'
                                 .format(node_idx))
            except ValueError:
                raise ValueError('Cannot guess the element of atom {}: '
                                 'the atom name has no alphabetic charater.'
                                 .format(node_idx))
            node['element'] = element


def categorical_cartesian_product(graph1, graph2, attributes=tuple()):
    product = nx.Graph()  # FIXME graphtype?
    for idx1, idx2 in itertools.product(graph1, graph2):
        node1 = graph1.nodes[idx1]
        node2 = graph2.nodes[idx2]
        if all((attr in node1 and attr in node2 and node1[attr] == node2[attr])
               or (attr not in node1 and attr not in node2) for attr in attributes):
            attrs = {}
            for attr in set(node1.keys()) | set(node2.keys()):
                attrs[attr] = (node1.get(attr, None), node2.get(attr, None))
            product.add_node((idx1, idx2), **attrs)
    return product


def categorical_modular_product(graph1, graph2, attributes=tuple()):
    product = categorical_cartesian_product(graph1, graph2, attributes)
    for (graph1_node1, graph2_node1), (graph1_node2, graph2_node2) in itertools.combinations(product.nodes(), 2):
        graph1_nodes = graph1_node1, graph1_node2
        graph2_nodes = graph2_node1, graph2_node2
        both_edge = graph1.has_edge(*graph1_nodes) and graph2.has_edge(*graph2_nodes)
        neither_edge = not graph1.has_edge(*graph1_nodes) and\
                       not graph2.has_edge(*graph2_nodes)
        # Effectively: not (graph1.has_edge(graph1_node1, graph1_node2) xor graph2.has_edge(graph2_node1, graph2_node2))
        if graph1_node1 != graph1_node2 and graph2_node1 != graph2_node2 and\
                (both_edge or neither_edge):
            attrs = {}
            if both_edge:
                g1_edge_keys = set(graph1.edges[graph1_nodes].keys())
                g2_edge_keys = set(graph2.edges[graph2_nodes].keys())
                for attr in g1_edge_keys | g2_edge_keys:
                    attrs[attr] = (graph1.edges[graph1_nodes].get(attr, None),
                                   graph2.edges[graph2_nodes].get(attr, None))
            product.add_edge((graph1_node1, graph2_node1), (graph1_node2, graph2_node2), **attrs)
    return product


def categorical_maximum_common_subgraph(graph1, graph2, attributes=tuple()):
    product = categorical_modular_product(graph1, graph2, attributes)
    cliques = nx.find_cliques(product)
    # cliques is an iterator which will return a *lot* of items. So make sure
    # we never turn it into a full list.
    largest = maxes(cliques, key=len)
    matches = [dict(clique) for clique in largest]
    return matches


def rate_match(residue, bead, match):
    """
    A helper function which rates how well ``match`` describes the isomorphism
    between ``residue`` and ``bead`` based on the number of matching atomnames.


    Parameters
    ----------
    residue : networkx.Graph
        A graph. Required node attributes:

            :atomname: The name of an atom.

    bead : networkx.Graph
        A subgraph of ``residue`` where the isomorphism is described by ``match``.
        Required node attributes:

            :atomname: The name of an atom.


    Returns
    -------
    int
        The number of entries in match where the atomname in ``residue`` matches
        the atomname in ``bead``.
    """
    return sum(residue.nodes[rdx].get('atomname') == bead.nodes[bdx].get('atomname')
               for rdx, bdx in match.items())


def _items_with_common_values(graph, nodes=None, excluded_keys=[]):
    """
    Finds all node attributes of nodes in graph that all have the same values.
    Returns a dict of node attribute/common value pairs. One can exclude specifc
    keys using the exclude variable.

    Parameters
    ----------
    graph: networkx.Graph
    nodes: collections.abc.Iterable or None
        If None, consider all nodes. All nodes must be in graph.
    exclude:  collections.abc.Iterable
        keys to exclude from the common value list

    Returns
    -------
    dict
    """
    if not nodes:
        nodes = set(graph.nodes)

    common_attrs = defaultdict(list)
    for idx in nodes:
        for key, val in graph.nodes[idx].items():
            if key not in excluded_keys:
                common_attrs[key].append(val)
    common_attrs = {key: vals[0] for key, vals in common_attrs.items()
                    if len(vals) == len(nodes) and are_all_equal(vals)}
    return common_attrs


def get_attrs(node, attrs):
    """
    Returns multiple values from a dictionary in order.

    Parameters
    ----------
    node: dict
        The dict from which items should be taken.
    attrs: collections.abc.Iterable
        The keys which values should be taken.

    Returns
    -------
    tuple
        A tuple containing the value of every key in attrs in the same order,
        where missing values are `None`.
    """
    return tuple(node.get(attr) for attr in attrs)


def partition_graph(graph, partitions):
    """
    Create a new graph based on `graph`, where nodes are aggregated based on
    `partitions`, similar to :func:`~networkx.algorithms.minors.quotient_graph`,
    except that it only accepts pre-made partitions, and edges are not given
    a 'weight' attribute. Much fast than the quotient_graph, since it creates
    edges based on existing edges rather than trying all possible combinations.

    Parameters
    ----------
    graph: networkx.Graph
        The graph to partition
    partitions: collections.abc.Iterable[collections.abc.Iterable[collections.abc.Hashable]]
        E.g. a list of lists of node indices, describing the partitions. Will
        be sorted by lowest index.

    Returns
    -------
    networkx.Graph
        The coarser graph.
    """
    new_graph = nx.Graph()
    partitions = sorted(partitions, key=min)
    mapping = {}
    for idx, node_idxs in enumerate(partitions):
        subgraph = nx.subgraph(graph, node_idxs)
        new_graph.add_node(idx,
                           graph=nx.subgraph(graph, node_idxs),
                           nnodes=len(subgraph),
                           nedges=len(subgraph.edges),
                           density=nx.density(subgraph))
        mapping.update({node_idx: idx for node_idx in node_idxs})

    for idx, jdx in graph.edges:
        if mapping[idx] != mapping[jdx]:
            new_idx, new_jdx = mapping[idx], mapping[jdx]
            new_graph.add_edge(new_idx, new_jdx)
    return new_graph


def make_residue_graph(graph, attrs=('chain', 'resid', 'resname', 'insertion_code')):
    """
    Create a new graph based on `graph`, where nodes with identical attribute
    values for the attribute names in `attrs` will be contracted into a single,
    coarser node. With the default arguments it will create a graph with one
    node per residue.
    Resulting (coarse) nodes will have the same attributes as the constructing
    nodes, but only those that have identical values. In addition, they'll have
    attributes 'graph', 'nnodes', 'nedges' and 'density'.

    Parameters
    ----------
    graph: networkx.Graph
        The graph to condense.
    attrs: collections.abc.Iterable[collections.abc.Hashable]
        The node attributes that determine node equivalence.

    Returns
    -------
    networkx.Graph
        The resulting coarser graph, where equivalent nodes are contracted to a
        single node.
    """
    # Create partitions. These will contain all nodes, even those without e.g.
    # a resname, since those will get resname None
    residue_idxs = collect_residues(graph, attrs)
    res_graph = partition_graph(graph, residue_idxs.values())
    # Alternatively we can use nx.quotient_graph, but that slows down the
    # program for large-ish molecules, since quotient_graph scales as O(N^2).
    # Our partition_graph scales much better, as O(E*N) (number of edges, nodes)
    # res_graph = nx.quotient_graph(graph,
    #                               sorted(residue_idxs.values(), key=min),
    #                               relabel=True)
    # Using this equivalence function rather than the preformed partitions
    # Creates an equivalent graph, but the node indices are numbered
    # differently. At the very least it would require a change in the tests.
    # Note2: This would probably break e.g. the DSSP processor, because the
    # ordering of residues in the molecule comes from here, and is used by
    # molecule.iter_residues.
    # def node_equiv(idx, jdx):
    #     return get_attrs(graph.nodes[idx], attrs) == get_attrs(graph.nodes[jdx], attrs)
    # res_graph = nx.quotient_graph(graph, node_equiv, relabel=True)
    for res_idx in res_graph:
        res_node = res_graph.nodes[res_idx]
        res_node.update(_items_with_common_values(res_node['graph'], excluded_keys=['graph']))
    return res_graph


def collect_residues(graph, attrs=('chain', 'resid', 'resname', 'insertion_code')):
    """
    Creates groups of indices based on the node attributes with keys `attrs`.
    All nodes in graph will be part of exactly one group.

    Parameters
    ----------
    graph: :class:`networkx.Graph`
        The graph whose node indices should be grouped.
    attrs: :class:`~collections.abc.Sequence`
        The attribute keys that should be used to group node indices. The
        associated values should be hashable.

    Returns
    -------
    dict[tuple, set]
        The keys are the found node attributes, the values the associated node
        indices.
    """
    residues = defaultdict(set)
    for node_idx in graph:
        key = get_attrs(graph.nodes[node_idx], attrs=attrs)
        residues[key].add(node_idx)
    return dict(residues)


# We can't inherit from nx.isomorphism.GraphMatcher to override
# `semantic_feasibility`. That implementation will clobber this one's method.
class MappingGraphMatcher(nx.isomorphism.isomorphvf2.GraphMatcher):
    def __init__(self, *args, edge_match=None, node_match=None, **kwargs):
        self.edge_match = edge_match
        self.node_match = node_match
        super().__init__(*args, **kwargs)
        self.G1_adj = self.G1.adj

    def semantic_feasibility(self, G1_node, G2_node):
        """
        Returns True if mapping G1_node to G2_node is semantically feasible.
        Adapted from networkx.algorithms.isomorphism.vf2userfunc._semantic_feasibility.
        """
        # Make sure the nodes match
        if self.node_match is not None:
            nm = self.node_match(self.G1.nodes[G1_node], self.G2.nodes[G2_node])
            if not nm:
                return False

        # Make sure the edges match
        if self.edge_match is not None:

            # Cached lookups
            core_1 = self.core_1
            edge_match = self.edge_match

            for neighbor in self.G1_adj[G1_node]:
                # G1_node is not in core_1, so we must handle R_self separately
                if neighbor == G1_node:
                    if not edge_match(G1_node, G1_node, G2_node, G2_node):
                        return False
                elif neighbor in core_1:
                    if not edge_match(G1_node, neighbor, G2_node, core_1[neighbor]):
                        return False
            # syntactic check has already verified that neighbors are symmetric
        return True



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides functionality to read and write GRO96 files.
"""

from itertools import chain

import numpy as np

from ..file_writer import open
from ..molecule import Molecule
from ..truncating_formatter import TruncFormatter
from ..utils import first_alpha


def read_gro(file_name, exclude=('SOL',), ignh=False):
    """
    Parse a gro file to create a molecule.

    Parameters
    ----------
    filename: str
        The file to read.
    exclude: collections.abc.Container[str]
        Atoms that have one of these residue names will not be included.
    ignh: bool
        Whether hydrogen atoms should be ignored.

    Returns
    -------
    vermouth.molecule.Molecule
        The parsed molecules. Will not contain edges.
    """
    molecule = Molecule()
    idx = 0
    field_types = [int, str, str, int, float, float, float]
    field_names = ['resid', 'resname', 'atomname', 'atomid', 'x', 'y', 'z']
    field_widths = [5, 5, 5, 5]

    with open(str(file_name)) as gro:
        next(gro)  # skip title
        num_atoms = int(next(gro))

        # We need the first line to figure out the exact format. In particular,
        # the precision and whether it has velocities.
        first_line = next(gro)
        has_vel = first_line.count('.') == 6
        first_dot = first_line.find('.', 25)
        second_dot = first_line.find('.', first_dot + 1)
        precision = second_dot - first_dot

        field_widths.extend([precision] * 3)
        if has_vel:
            field_widths.extend([precision] * 3)
            field_types.extend([float] * 3)
            field_names.extend(['vx', 'vy', 'vz'])

        start = 0
        slices = []
        for width in field_widths:
            if width > 0:
                slices.append(slice(start, start + width))
            start = start + abs(width)

        # Start parsing the file in earnest. And let's not forget the first
        # line.
        for line_idx, line in enumerate(chain([first_line], gro)):
            properties = {}
            # This (apart maybe from adhering to the number of lines specified
            # by the file) is the fastest method of checking whether we are at
            # the last line (box) of the file. Other things tested: regexp
            # matching, looking ahead, and testing whether the line looks like
            # a box-line. I think the reason this is faster is because the try
            # block will almost never raise an exception.
            try:
                for name, type_, slice_ in zip(field_names, field_types, slices):
                    properties[name] = type_(line[slice_].strip())
            except ValueError:
                if line_idx != num_atoms:
                    raise
                break

            properties['element'] = first_alpha(properties['atomname'])
            properties['chain'] = ''
            if properties['resname'] in exclude or (ignh and properties['element'] == 'H'):
                continue

            pos = (properties.pop('x'), properties.pop('y'), properties.pop('z'))
            properties['position'] = np.array(pos, dtype=float)

            if has_vel:
                vel = (properties.pop('vx'), properties.pop('vy'), properties.pop('vz'))
                properties['velocity'] = np.array(vel, dtype=float)

            molecule.add_node(idx, **properties)
            idx += 1
    return molecule


def write_gro(system, file_name, precision=7, title='Martinized!', box=(0, 0, 0)):
    """
    Write `system` to `file_name`, which will be a GRO96 file.

    Parameters
    ----------
    system: vermouth.system.System
        The system to write.
    file_name: str
        The file to write to.
    precision: int
        The desired precision for coordinates and (optionally) velocities.
    title: str
        Title for the gro file.
    box: tuple[float]
        Box length and optionally angles.
    """
    formatter = TruncFormatter()
    pos_format_string = '{{:{ntx}.3ft}}'.format(ntx=precision + 1)
    format_string = '{:5dt}{:<5st}{:>5st}{:5dt}' + pos_format_string*3
    # Pick an arbitrary node from the first molecule to see if all molecules
    # have velocities. Somehow I don't think we can write velocities for some
    # molecules but not others...
    has_vel = all('velocity' in next(iter(mol.nodes.values())) for mol in system.molecules)
    if has_vel:
        vel_format_string = '{{:{ntx}.4ft}}'*3
        vel_format_string = vel_format_string.format(ntx=precision+1)

    with open(str(file_name), 'w') as out:
        out.write(title + '\n')  # Title
        out.write(formatter.format('{}\n', system.num_particles))  # number of atoms
        atomid = 1
        for molecule in system.molecules:
            node_order = molecule.nodes
            for node_idx in node_order:
                node = molecule.nodes[node_idx]
                atomname = node['atomname']
                resname = node['resname']
                resid = node['resid']
                x, y, z = node['position']  # pylint: disable=invalid-name

                line = formatter.format(format_string, resid, resname, atomname,
                                        atomid, x, y, z)
                if has_vel:
                    vx, vy, vz = node['velocity']  # pylint: disable=invalid-name
                    line += formatter.format(vel_format_string, vx, vy, vz)
                atomid += 1
                out.write(line + '\n')
        # Box
        out.write(' '.join(str(value) for value in box))
        # to appease VMD which cannot read the file otherwise.
        out.write('\n')



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Contains unittests for the functions in vermouth.graph_utils.
"""


from pprint import pprint

import networkx as nx
import pytest
import vermouth
from .helper_functions import make_into_set

# no-member because module networkx does indeed have a member isomorphism;
# too-many-arguments because some tests just need a lot of data.
# pylint: disable=no-member, too-many-arguments


def basic_molecule(node_data, edge_data=None):
    """
    Construct a simple Molecule based with specified nodes and edges.
    """
    if edge_data is None:
        edge_data = {}

    mol = vermouth.Molecule()
    for idx, node in enumerate(node_data):
        mol.add_node(idx, **node)
    for (idx, jdx), data in edge_data.items():
        mol.add_edge(idx, jdx, **data)
    return mol


@pytest.mark.parametrize('node_data_in, expected_node_data', [
    ([], []),
    ([{'atomname': 'H3'}],
     [{'atomname': 'H3', 'element': 'H'}]),
    ([{'atomname': '1H3'}],
     [{'atomname': '1H3', 'element': 'H'}]),
    ([{'atomname': 'H3'}, {'atomname': '1H3'}],
     [{'atomname': 'H3', 'element': 'H'}, {'atomname': '1H3', 'element': 'H'}]),
    ([{'atomname': 'Cl1', 'element': 'Cl', 'attr': None},
      {'atomname': '31C3'}],
     [{'atomname': 'Cl1', 'element': 'Cl', 'attr': None},
      {'atomname': '31C3', 'element': 'C'}]),
    ([{'element': 'Cl'}, {'atomname': '31C3'}],
     [{'element': 'Cl'}, {'atomname': '31C3', 'element': 'C'}]),
])
def test_add_element_attr(node_data_in, expected_node_data):
    """
    Tests for the function ``add_element_attr``.
    """
    mol = basic_molecule(node_data_in)
    vermouth.graph_utils.add_element_attr(mol)
    expected = basic_molecule(expected_node_data)
    assert mol.nodes(data=True) == expected.nodes(data=True)


@pytest.mark.parametrize('node_data_in, exception', [
    ([{'atomname': '1234'}], ValueError),  # No alpha character
    ([{'peanuts': '1H3'}], ValueError),  # Not atomname attribute
    ([{'atomname': 'H3'}, {'atomname': '1234'}], ValueError),  # No alpha character for second atom
    ([{'atomname': 'H3'}, {'peanuts': '1234'}], ValueError),  # No atomname attribute for second atom
])
def test_add_element_attr_errors(node_data_in, exception):
    """
    Tests to make sure the function ``add_element_attr`` raises the expected
    errors.
    """
    mol = basic_molecule(node_data_in)
    with pytest.raises(exception):
        vermouth.graph_utils.add_element_attr(mol)


@pytest.mark.parametrize('node_data1, node_data2, attrs, expected', [
    ([], [], [], {}),
    ([{}], [{}], [], {(0, 0): {}}),
    (
        [{'a': 1}],
        [{'a': 2}],
        [],
        {(0, 0): {'a': (1, 2)}}
    ),
    (
        [{'a': 1}],
        [{'a': 1}],
        ['a'],
        {(0, 0): {'a': (1, 1)}}
    ),
    (
        [{'a': 1}],
        [{'a': 2}],
        ['a'],
        {}
    ),
    (
        [{'a': 1}],
        [{'b': 1}],
        ['a'],
        {}
    ),
    (
        [{'a': 1, 'b': 2}],
        [{'a': 1, 'b': 1}],
        ['a'],
        {(0, 0): {'a': (1, 1), 'b': (2, 1)}}
    ),
    (
        [{"1": 1, "2": 2}],
        [{"2": 2, "3": 3}],
        ["2"],
        {(0, 0): {"1": (1, None), "2": (2, 2), "3": (None, 3)}}
    ),
    (
        [{'a': 1, 'b': 2}],
        [{'a': 1, 'b': 1}],
        ['a', 'b'],
        {}
    ),
    (
        [{'a': 1}, {'b': 2}],
        [{'b': 2}, {'c': 3}],
        ['b'],
        {(1, 0): {'b': (2, 2)}, (0, 1): {'a': (1, None), 'c': (None, 3)}}
    ),
    (
        [{'a': 1, 'b': 1}, {'b': 2}],
        [{'b': 2}, {'c': 3, 'b': 2}],
        ['b'],
        {(1, 0): {'b': (2, 2)}, (1, 1): {'b': (2, 2), 'c': (None, 3)}}
    ),
    (
        [{'a': 1, 'b': 1}, {'b': 2, 'a': 2}],
        [{'a': 1, 'b': 1}, {'a': 2, 'b': 2}],
        ['a'],
        {(0, 0): {'a': (1, 1), 'b': (1, 1)}, (1, 1): {'a': (2, 2), 'b': (2, 2)}}
    ),
])
def test_categorical_cartesian_product(node_data1, node_data2, attrs, expected):
    """
    Tests for the function ``categorical_cartesian_product``.
    """
    graph1 = basic_molecule(node_data1)
    graph2 = basic_molecule(node_data2)
    found = vermouth.graph_utils.categorical_cartesian_product(graph1, graph2, attrs)
    expected_mol = nx.Graph()

    # Only test nodes, because the categorical product does not contain edges.
    for idx, data in expected.items():
        expected_mol.add_node(idx, **data)
    assert expected_mol.nodes(data=True) == found.nodes(data=True)


@pytest.mark.parametrize('node_data1, edges1, node_data2, edges2, attrs, expected_nodes, expected_edges', [
    ([], {}, [], {}, [], {}, {}),
    (
        [{}, {}], {},
        [{}, {}], {(0, 1): {}},
        [],
        {(0, 0): {}, (1, 0): {}, (0, 1): {}, (1, 1): {}},
        {}
    ),
    (
        [{'a': 1}, {'a': 2}], {},
        [{'a': 1}, {'a': 2}], {},
        ['a'],
        {(0, 0): {'a': (1, 1)}, (1, 1): {'a': (2, 2)}},
        {((0, 0), (1, 1)): {}}
    ),
    (
        [{}, {}], {(0, 1): {}},
        [{}, {}], {(0, 1): {}},
        [],
        {(0, 0): {}, (1, 0): {}, (0, 1): {}, (1, 1): {}},
        {((0, 0), (1, 1)): {}, ((0, 1), (1, 0)): {}}
    ),
    (
        [{}, {}], {},
        [{}, {}], {},
        [],
        {(0, 0): {}, (1, 0): {}, (0, 1): {}, (1, 1): {}},
        {((0, 0), (1, 1)): {}, ((0, 1), (1, 0)): {}}
    ),
    (
        [{}, {}], {(0, 1): {'a': 1}},
        [{}, {}], {(0, 1): {'b': 1}},
        [],
        {(0, 0): {}, (1, 0): {}, (0, 1): {}, (1, 1): {}},
        {((0, 0), (1, 1)): {'a': (1, None), 'b': (None, 1)},
         ((0, 1), (1, 0)): {'a': (1, None), 'b': (None, 1)}}
    ),
])
def test_categorical_modular_product(node_data1, edges1, node_data2, edges2,
                                     attrs, expected_nodes, expected_edges):
    """
    Tests for the function ``categorical_modular_product``.
    """
    graph1 = basic_molecule(node_data1, edges1)
    graph2 = basic_molecule(node_data2, edges2)
    found = vermouth.graph_utils.categorical_modular_product(graph1, graph2, attrs)
    expected_mol = nx.Graph()
    for idx, data in expected_nodes.items():
        expected_mol.add_node(idx, **data)
    for edge_idx, data in expected_edges.items():
        expected_mol.add_edge(*edge_idx, **data)

    assert found.nodes(data=True) == expected_mol.nodes(data=True)
    edges_seen = set()
    for idx, jdx, data in found.edges(data=True):
        assert expected_mol.edges[idx, jdx] == data
        edges_seen.add(frozenset((idx, jdx)))
    assert set(frozenset(edge) for edge in expected_mol.edges) == edges_seen


@pytest.mark.parametrize('node_data1, edges1, node_data2, edges2, attrs, expected', [
    ([], {}, [], {}, ['id'], []),
    (
        [{'id': 1}], {},
        [{'id': 1}], {},
        ['id'],
        [{0: 0}]
    ),
    (
        [{'id': 1}, {'id': 2}], {},
        [{'id': 1}, {'id': 2}], {},
        ['id'],
        [{0: 0, 1: 1}]
    ),
    (
        [{'id': 1}, {'id': 2}], {(0, 1): {}},
        [{'id': 1}, {'id': 2}], {},
        ['id'],
        [{0: 0}, {1: 1}]
    ),
    (
        [{'id': 0}, {'id': 1}, {'id': 2}], {(0, 1): {}, (1, 2): {}},
        [{'id': 1}, {'id': 2}, {'id': 3}], {(0, 1): {}, (1, 2): {}},
        ['id'],
        [{1: 0, 2: 1}]
    ),
    (
        [{}, {}], {(0, 1): {}},
        [{}, {}], {(0, 1): {}},
        [],
        [{0: 0, 1: 1}, {0: 1, 1: 0}]
    ),
])
def test_categorical_maximum_common_subgraph(node_data1, edges1, node_data2,
                                             edges2, attrs, expected):
    """
    Tests for the function ``categorical_maximum_common_subgraph``.
    """
    graph1 = basic_molecule(node_data1, edges1)
    graph2 = basic_molecule(node_data2, edges2)
    found = vermouth.graph_utils.categorical_maximum_common_subgraph(graph1, graph2, attrs)
    assert make_into_set(found) == make_into_set(expected)


@pytest.mark.parametrize('nodes1, nodes2, match, expected', [
    ([], [], {}, 0),
    (
        [{'atomname': 0}],
        [{'atomname': 1}],
        {0: 0}, 0
    ),
    (
        [{'atomname': 0}],
        [{'atomname': 1}],
        {}, 0
    ),
    (
        [{'atomname': 0}],
        [{'atomname': 0}],
        {}, 0
    ),
    (
        [{'atomname': 0}],
        [{'atomname': 0}],
        {0: 0}, 1
    ),
    (
        [{'atomname': 0}, {'atomname': 0}],
        [{'atomname': 0}, {'atomname': 0}],
        {0: 0}, 1
    ),
    (
        [{'atomname': 0}, {'atomname': 0}],
        [{'atomname': 0}, {'atomname': 0}],
        {0: 0, 1: 1}, 2
    ),
    (
        [{'atomname': 0}, {'atomname': 0}],
        [{'atomname': 0}, {'atomname': 1}],
        {0: 0, 1: 1}, 1
    ),
    (
        [{'atomname': 0}, {'atomname': 1}],
        [{'atomname': 0}, {'atomname': 1}],
        {0: 0, 1: 1}, 2
    ),
    (
        [{'atomname': 0}, {'atomname': 1}],
        [{'atomname': 0}, {'atomname': 1}],
        {0: 1}, 0
    ),
])
def test_rate_match(nodes1, nodes2, match, expected):
    """
    Tests for the function ``rate_match``.
    """
    mol1 = basic_molecule(nodes1)
    mol2 = basic_molecule(nodes2)
    found = vermouth.graph_utils.rate_match(mol1, mol2, match)
    assert found == expected


@pytest.mark.parametrize('nodes1, edges1, nodes2, edges2', [
    ([], {}, [], {}),
    (
        [{'chain': 0, 'resid': 0, 'resname': 0}], {},
        [{'chain': 0, 'resid': 0, 'resname': 0}], {}
    ),
    (
        [{'chain': 0, 'resid': 0, 'resname': 1}], {},
        [{'chain': 0, 'resid': 0, 'resname': 1}], {}
    ),
    (
        [{'chain': 0, 'resid': 2, 'resname': 1}], {},
        [{'chain': 0, 'resid': 2, 'resname': 1}], {}
    ),
    (
        [{'chain': 0, 'resid': 2, 'resname': 1, 'attr': 5}], {},
        [{'chain': 0, 'resid': 2, 'resname': 1, 'attr': 5}], {}
    ),
    (
        [{'chain': 0, 'resid': 2, 'resname': 1, 'attr': 5},
         {'chain': 0, 'resid': 2, 'resname': 1, 'attr': 7}],
        {},
        [{'chain': 0, 'resid': 2, 'resname': 1}],
        {}
    ),
    (
        [{'chain': 0, 'resid': 2, 'resname': 1, 'attr': 5},
         {'chain': 0, 'resid': 2, 'resname': 2, 'attr': 7}],
        {},
        [{'chain': 0, 'resid': 2, 'resname': 1, 'attr': 5},
         {'chain': 0, 'resid': 2, 'resname': 2, 'attr': 7}],
        {}
    ),
    (
        [{'chain': 0, 'resid': 2, 'resname': 1, 'attr': 5},
         {'chain': 0, 'resid': 2, 'resname': 2, 'attr': 7}],
        {(0, 1): {}},
        [{'chain': 0, 'resid': 2, 'resname': 1, 'attr': 5},
         {'chain': 0, 'resid': 2, 'resname': 2, 'attr': 7}],
        {(0, 1): {}}
    ),
    (
        [{'chain': 0, 'resid': 2, 'resname': 1, 'attr': 5},
         {'chain': 0, 'resid': 2, 'resname': 1, 'attr': 6},
         {'chain': 0, 'resid': 2, 'resname': 2, 'attr': 7}],
        {(2, 1): {}},
        [{'chain': 0, 'resid': 2, 'resname': 1},
         {'chain': 0, 'resid': 2, 'resname': 2, 'attr': 7}],
        {(0, 1): {}}
    ),
])
def test_make_residue_graph(nodes1, edges1, nodes2, edges2):
    """
    Tests for the function ``make_residue_graph``.
    """
    mol1 = basic_molecule(nodes1, edges1)
    found = vermouth.graph_utils.make_residue_graph(mol1)
    expected = basic_molecule(nodes2, edges2)

    for node in found:
        data = found.nodes[node]
        subgraph = data['graph']
        assert len(subgraph.nodes) == data['nnodes']
        assert len(subgraph.edges) == data['nedges']
        assert nx.density(subgraph) == data['density']
        del found.nodes[node]['graph']
        del found.nodes[node]['nnodes']
        del found.nodes[node]['nedges']
        del found.nodes[node]['density']
    pprint(("Found nodes", found.nodes(data=True)))
    pprint(("Expected nodes", expected.nodes(data=True)))

    assert found.nodes(data=True) == expected.nodes(data=True)
    edges_seen = set()
    for idx, jdx, data in found.edges(data=True):
        assert expected.has_edge(idx, jdx) and expected.edges[idx, jdx] == data
        edges_seen.add(frozenset((idx, jdx)))
    assert set(frozenset(edge) for edge in expected.edges) == edges_seen

@pytest.mark.parametrize('nodes, attrs, expected', [
    ([], [], {}),
    ([], [1], {}),
    ([{}, {}], [], {tuple(): {0, 1}}),
    ([{1: 1}, {1: 1}], [], {tuple(): {0, 1}}),
    ([{1: 1}, {1: 1}], [1], {(1,): {0, 1}}),
    ([{1: 1}, {1: 2}], [1], {(1,): {0}, (2,): {1}}),
    ([{1: 1}, {1: 1, 2: 2}], [1], {(1,): {0, 1}}),
    ([{1: 1, 2: 1}, {1: 1, 2: 2}], [1], {(1,): {0, 1}}),
    ([{1: 1, 2: 1}, {1: 1, 2: 2}], [1, 2], {(1, 1): {0}, (1, 2): {1}}),
    ([{1: 1}, {1: 1, 2: 2}], [1, 2], {(1, None): {0}, (1, 2): {1}}),

])
def test_collect_residues(nodes, attrs, expected):
    """
    Tests for ``collect_residues``
    """
    graph = nx.Graph()
    graph.add_nodes_from(enumerate(nodes))
    found = vermouth.graph_utils.collect_residues(graph, attrs)
    assert found == expected



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Test the functions required to use DSSP.
"""

import os
import glob
import itertools

import numpy as np
import pytest

import vermouth
from vermouth.file_writer import DeferredFileWriter
from vermouth.forcefield import get_native_force_field
from vermouth.dssp import dssp
from vermouth.dssp.dssp import DSSPError
from vermouth.pdb.pdb import read_pdb
from vermouth.tests.datafiles import (
    PDB_PROTEIN,
    DSSP_OUTPUT,
    PDB_ALA5_CG,
)

DSSP_EXECUTABLE = os.environ.get('VERMOUTH_TEST_DSSP', 'dssp')
SECSTRUCT_1BTA = list('CEEEEETTTCCSHHHHHHHHHHHHTCCTTCCCSHHHHHHHHTTT'
                      'SCSSEEEEEESTTHHHHTTTSSHHHHHHHHHHHHHTTCCEEEEEC')


# TODO: The code is very repetitive. There may be a way to refactor it with
# clever use of parametrize and fixtures.
class TestAnnotateResidues:
    """
    Tests for the :class:`dssp.AnnotateResidues` processor.
    """
    @staticmethod
    def build_molecule(nresidues):
        """
        Build a dummy molecule with the requested number of residues and 3
        atoms per residue.
        """
        molecule = vermouth.molecule.Molecule()
        residue_template = vermouth.molecule.Molecule()
        residue_template.add_nodes_from(
            (idx, {'chain':'', 'atomname':str(idx), 'resname': 'DUMMY', 'resid': 1})
            for idx in range(3)
        )
        for _ in range(nresidues):
            molecule.merge_molecule(residue_template)
        return molecule

    @staticmethod
    def sequence_from_mol(molecule, attribute, default=None):
        """
        Extract the content of an attribute for each node of a molecule.
        """
        return [node.get(attribute, default) for node in molecule.nodes.values()]

    def sequence_from_system(self, system, attribute, default=None):
        """
        Extract the content of an attribute for each node of a system.
        """
        return list(itertools.chain(*(
            self.sequence_from_mol(molecule, attribute, default)
            for molecule in system.molecules
        )))

    @pytest.mark.parametrize('nres', (0, 1, 3, 10))
    def test_build_molecule(self, nres):
        """
        :meth:`build_molecule` and :meth:`sequence_from_mol` work as excpected.
        """
        expected_resid = list(itertools.chain(
            *([idx + 1] * 3 for idx in range(nres))
        ))
        expected_atomname = ['0', '1', '2'] * nres
        expected_chain = [''] * (nres * 3)
        expected_resname = ['DUMMY'] * (nres * 3)
        molecule = self.build_molecule(nres)
        assert self.sequence_from_mol(molecule, 'resid') == expected_resid
        assert self.sequence_from_mol(molecule, 'resname') == expected_resname
        assert self.sequence_from_mol(molecule, 'chain') == expected_chain
        assert self.sequence_from_mol(molecule, 'atomname') == expected_atomname

    @pytest.fixture
    def single_mol_system(self):
        """
        Build a system with a single molecule that count 5 residues of 3 atoms
        each.
        """
        molecule = self.build_molecule(5)
        system = vermouth.system.System()
        system.molecules = [molecule]
        return system

    @pytest.fixture
    def multi_mol_system_irregular(self):
        """
        Build a system with 3 molecules having 4, 5, and 6 residues,
        respectively.
        """
        system = vermouth.system.System()
        system.molecules = [self.build_molecule(nres) for nres in (4, 5, 6)]
        return system

    @pytest.fixture
    def multi_mol_system_regular(self):
        """
        Build a system with 3 molecules having each 4 residues.
        """
        system = vermouth.system.System()
        system.molecules = [self.build_molecule(4) for _ in range(3)]
        return system

    @pytest.mark.parametrize('sequence', (
        'ABCDE',
        ['A', 'B', 'C', 'D', 'E'],
        range(5),
    ))
    def test_single_molecule(self, single_mol_system, sequence):
        """
        The simple case with a single molecule and a sequence of the right size
        works as expected.
        """
        expected = list(itertools.chain(
            *([element] * 3 for element in sequence)
        ))
        processor = dssp.AnnotateResidues('test', sequence)
        processor.run_system(single_mol_system)
        found = self.sequence_from_system(single_mol_system, 'test')
        assert found == expected

    @pytest.mark.parametrize('sequence', (
        'ABCDEFGHIJKLMNO',
        list('ABCDEFGHIJKLMNO'),
        range(15),
    ))
    def test_multi_molecules_diff_sizes(self, multi_mol_system_irregular, sequence):
        """
        The case of many protein of various sizes and a sequence of the right
        size works as expected.
        """
        expected = list(itertools.chain(
            *([element] * 3 for element in sequence)
        ))
        processor = dssp.AnnotateResidues('test', sequence)
        processor.run_system(multi_mol_system_irregular)
        found = self.sequence_from_system(multi_mol_system_irregular, 'test')
        assert found == expected

    @pytest.mark.parametrize('sequence', (
        'ABCD',
        ['A', 'B', 'C', 'D'],
        range(4),
    ))
    def test_multi_molecules_cycle(self, multi_mol_system_regular, sequence):
        """
        The case with multiple molecules with all the same size and one
        sequence to repeat for each molecule works as expected.
        """
        expected = list(itertools.chain(
            *([element] * 3 for element in sequence)
        ))
        expected = expected * 3
        processor = dssp.AnnotateResidues('test', sequence)
        processor.run_system(multi_mol_system_regular)
        found = self.sequence_from_system(multi_mol_system_regular, 'test')
        assert found == expected

    def test_single_molecules_cycle_one(self, single_mol_system):
        """
        One molecule and a one element sequence to repeat over all residues of
        the molecule.
        """
        sequence = 'A'
        expected = [sequence] * (5 * 3)
        processor = dssp.AnnotateResidues('test', sequence)
        processor.run_system(single_mol_system)
        found = self.sequence_from_system(single_mol_system, 'test')
        assert found == expected


    def test_multi_molecules_cycle_one(self, multi_mol_system_irregular):
        """
        Many molecules and a one element sequence to repeat.
        """
        sequence = 'A'
        expected = [sequence] * (15 * 3)
        processor = dssp.AnnotateResidues('test', sequence)
        processor.run_system(multi_mol_system_irregular)
        found = self.sequence_from_system(multi_mol_system_irregular, 'test')
        assert found == expected

    @staticmethod
    @pytest.mark.parametrize('sequence', (
        'ABC',  # Too short
        'ABCD',  # Too short, match the length of the first molecule
        'ABCDEFGHIFKLMNOPQRSTU',  # Too long
        '',  # Empty
    ))
    def test_wrong_length(multi_mol_system_irregular, sequence):
        """
        Many molecule and a sequence that has the wrong length raises an error.
        """
        processor = dssp.AnnotateResidues('test', sequence)
        with pytest.raises(ValueError):
            processor.run_system(multi_mol_system_irregular)

    @staticmethod
    @pytest.mark.parametrize('sequence', (
        'ABC',  # Too short
        'ABCD',  # Too short, match the length of the first molecule
        'ABCDEFGHIFKLMNOPQRSTU',  # Too long
        '',  # Empty
        'ABCDEFGHIJKLMNO',  # Length of all the molecules, without filter
    ))
    def test_wrong_length_with_filter(multi_mol_system_irregular, sequence):
        """
        Many molecules and a sequence that has the wrong length because of a
        molecule selector.
        """
        # We exclude the second molecule. The filter excludes it based on the
        # number of nodes, which is 15 because it has 5 residues with 3 nodes
        # each.
        processor = dssp.AnnotateResidues(
            'test', sequence,
            molecule_selector=lambda mol: len(mol.nodes) != (5 * 3),
        )
        with pytest.raises(ValueError):
            processor.run_system(multi_mol_system_irregular)

    @staticmethod
    def test_empty_system_empty_sequence():
        """
        There are no molecules, but the sequence is empty.
        """
        system = vermouth.system.System()
        sequence = ''
        processor = dssp.AnnotateResidues('test', sequence)
        try:
            processor.run_system(system)
        except ValueError:
            pytest.fail('Should not have raised a ValueError.')

    @staticmethod
    def test_empty_system_error():
        """
        There are no molecules, but there is a sequence. Should raise an error.
        """
        system = vermouth.system.System()
        sequence = 'not empty'
        processor = dssp.AnnotateResidues('test', sequence)
        with pytest.raises(ValueError):
            processor.run_system(system)

    @staticmethod
    def test_empty_with_filter(multi_mol_system_irregular):
        """
        There is a sequence, but no molecule are accepted by the molecule
        selector. Should raise an error.
        """
        sequence = 'not empty'
        processor = dssp.AnnotateResidues(
            'test', sequence, molecule_selector=lambda mol: False
        )
        with pytest.raises(ValueError):
            processor.run_system(multi_mol_system_irregular)

    def test_run_molecule(self, single_mol_system):
        """
        The `run_molecule` method works.
        """
        sequence = 'ABCDE'
        expected = list(itertools.chain(
            *([element] * 3 for element in sequence)
        ))
        processor = dssp.AnnotateResidues('test', sequence)
        processor.run_molecule(single_mol_system.molecules[0])
        found = self.sequence_from_system(single_mol_system, 'test')
        assert found == expected

    def test_run_molecule_not_selected(self, single_mol_system):
        """
        The molecule selector works with `run_molecule`.
        """
        sequence = 'ABCDE'
        processor = dssp.AnnotateResidues(
            'test', sequence, molecule_selector=lambda mol: False
        )
        processor.run_molecule(single_mol_system.molecules[0])
        found = self.sequence_from_system(single_mol_system, 'test')
        assert vermouth.utils.are_all_equal(found)
        assert found[0] is None


def test_read_dssp2():
    """
    Test that :func:`vermouth.dssp.dssp.read_dssp2` returns the expected
    secondary structure sequence.
    """
    with open(str(DSSP_OUTPUT)) as infile:
        secondary_structure = dssp.read_dssp2(infile)
    assert secondary_structure == SECSTRUCT_1BTA


@pytest.mark.parametrize('savefile', [True, False])
def test_run_dssp(savefile, tmpdir):
    """
    Test that :func:`vermouth.molecule.dssp.dssp.run_dssp` runs as expected and
    generate a save file only if requested.
    """
    # The test runs twice, once with the savefile set to True so we test with
    # saving the DSSP output to file, and once with savefile set t False so we
    # do not generate the file. The "savefile" argument is set by
    # pytest.mark.parametrize.
    # The "tmpdir" argument is set by pytest and is the path to a temporary
    # directory that exists only for one iteration of the test.
    if savefile:
        path = tmpdir.join('dssp_output')
    else:
        path = None
    system = vermouth.System()
    for molecule in read_pdb(str(PDB_PROTEIN)):
        system.add_molecule(molecule)
    secondary_structure = dssp.run_dssp(system,
                                        executable=DSSP_EXECUTABLE,
                                        savefile=path)

    # Make sure we produced the expected sequence of secondary structures
    assert secondary_structure == SECSTRUCT_1BTA

    # If we test with savefile, then we need to make sure the file is created
    # and its content corresponds to the reference (excluding the first lines
    # that are variable or contain non-essencial data read from the PDB file).
    # If we test without savefile, then we need to make sure the file is not
    # created.
    if savefile:
        DeferredFileWriter().write()
        assert path.exists()
        with open(str(path)) as genfile, open(str(DSSP_OUTPUT)) as reffile:
            # DSSP 3 is outputs mostly the same thing as DSSP2, though there
            # are some differences in non significant whitespaces, and an extra
            # field header. We need to normalize these differences to be able
            # to compare.
            gen = '\n'.join([
                line.strip().replace('            CHAIN', '')
                for line in genfile.readlines()[6:]
            ])
            ref = '\n'.join([line.strip() for line in reffile.readlines()[6:]])
            assert gen == ref
    else:
        # Is the directory empty?
        assert not os.listdir(str(tmpdir))


@pytest.mark.parametrize('pdb, loglevel,expected', [
    (PDB_PROTEIN, 10, True),  # DEBUG
    (PDB_PROTEIN, 30, False),  # WARNING
    # Using a CG pdb will cause a DSSP error, which should preserve the input
    (PDB_ALA5_CG, 10, True),  # DEBUG
    (PDB_ALA5_CG, 30, True),  # WARNING
])
def test_run_dssp_input_file(tmpdir, caplog, pdb, loglevel, expected):
    """
    Test that the DSSP input file is preserved (only) in the right conditions
    """
    caplog.set_level(loglevel)
    system = vermouth.System()
    for molecule in read_pdb(str(pdb)):
        system.add_molecule(molecule)
    with tmpdir.as_cwd():
        try:
            dssp.run_dssp(system, executable=DSSP_EXECUTABLE)
        except DSSPError:
            pass
        if expected:
            target = 1
        else:
            target = 0
        matches = glob.glob('dssp_in*.pdb')
        assert len(matches) == target, matches
        if matches:
            # Make sure it's a valid PDB file. Mostly anyway.
            list(read_pdb(matches[0]))


def test_cterm_atomnames():
    nodes = [
        dict(resname="ALA", atomname="N", element='N', resid=1, chain="", position=np.array([9.534, 5.359, 0.000])),
        dict(resname="ALA", atomname="CA", element='C', resid=1, chain="", position=np.array([10.190, 6.661, -0.000])),
        dict(resname="ALA", atomname="C", element='C', resid=1, chain="", position=np.array([11.706, 6.515, 0.000])),
        dict(resname="ALA", atomname="O", element='O', resid=1, chain="", position=np.array([12.232, 5.403, 0.000])),
        dict(resname="ALA", atomname="CB", element='C', resid=1, chain="", position=np.array([9.733, 7.484, 1.196])),
        dict(resname="ALA", atomname="H", element='H', resid=1, chain="", position=np.array([10.101, 4.523, 0.000])),
        dict(resname="ALA", atomname="HA", element='H', resid=1, chain="", position=np.array([9.914, 7.191, -0.912])),
        dict(resname="ALA", atomname="1HB", element='H', resid=1, chain="", position=np.array([10.231, 8.454, 1.181])),
        dict(resname="ALA", atomname="2HB", element='H', resid=1, chain="", position=np.array([8.654, 7.630, 1.147])),
        dict(resname="ALA", atomname="3HB", element='H', resid=1, chain="", position=np.array([9.987, 6.960, 2.116])),

        dict(resname="ALA", atomname="N", element='N', resid=2, chain="", position=np.array([12.404, 7.646, 0.000])),
        dict(resname="ALA", atomname="CA", element='C', resid=2, chain="", position=np.array([13.862, 7.646, 0.000])),
        dict(resname="ALA", atomname="C", element='C', resid=2, chain="", position=np.array([14.413, 9.066, 0.000])),
        dict(resname="ALA", atomname="O1", element='O', resid=2, chain="", position=np.array([14.462, 9.691, 1.023])),
        dict(resname="ALA", atomname="O2", element='O', resid=2, chain="", position=np.array([14.798, 9.560, -1.023])),
        dict(resname="ALA", atomname="CB", element='C', resid=2, chain="", position=np.array([14.392, 6.868, -1.196])),
        dict(resname="ALA", atomname="H", element='H', resid=2, chain="", position=np.array([11.912, 8.528, -0.000])),
        dict(resname="ALA", atomname="HA", element='H', resid=2, chain="", position=np.array([14.212, 7.162, 0.912])),
        dict(resname="ALA", atomname="1HB", element='H', resid=2, chain="", position=np.array([15.482, 6.878, -1.181])),
        dict(resname="ALA", atomname="2HB", element='H', resid=2, chain="", position=np.array([14.038, 5.839, -1.147])),
        dict(resname="ALA", atomname="3HB", element='H', resid=2, chain="", position=np.array([14.038, 7.331, -2.116])),

    ]
    edges = [
        (0, 1), (0, 5),
        (1, 2), (1, 4), (1, 6),
        (2, 3),
        (4, 7), (4, 8), (4, 9),
        (2, 10),
        (10, 11), (10, 16),
        (11, 12), (11, 15), (11, 17),
        (12, 13), (12, 14),
        (15, 18), (15, 19), (15, 20),
    ]
    ff = get_native_force_field('universal')
    mol = vermouth.molecule.Molecule(force_field=ff)
    mol.add_nodes_from(enumerate(nodes))
    mol.add_edges_from(edges)
    system = vermouth.system.System(force_field=ff)
    system.add_molecule(mol)
    vermouth.processors.RepairGraph().run_system(system)
    vermouth.processors.CanonicalizeModifications().run_system(system)
    dssp_out = dssp.run_dssp(system, executable=DSSP_EXECUTABLE)
    assert dssp_out == list('CC')



# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Test that force field files are properly read.
"""

import textwrap
import pytest
import vermouth.forcefield
import vermouth.molecule
import vermouth.gmx.itp_read

class TestITP:
    @staticmethod
    def test_moleculetype():
        lines = """
        [ moleculetype ]
        GLY  3
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.gmx.itp_read.read_itp(lines, ff)
        assert 'GLY' in ff.blocks
        assert ff.blocks['GLY'].nrexcl == 3

    @staticmethod
    def test_multiple_moleculetype():
        lines = """
            [ moleculetype ]
            GLY  3

            [ moleculetype ]
            VAL 2
            """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.gmx.itp_read.read_itp(lines, ff)
        assert set(ff.blocks) == set(['GLY', 'VAL'])
        assert ff.blocks['GLY'].nrexcl == 3
        assert ff.blocks['VAL'].nrexcl == 2

    @staticmethod
    def test_atoms():
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4  1  ALA  BB  1
        2 P3\t1\tALA  SC1 2 -3
        3 P2  1  ALA  SC2 3 -3 72
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.gmx.itp_read.read_itp(lines, ff)
        block = ff.blocks['GLY']
        assert len(block.nodes) == 3
        assert block.nodes[0] == {'index': 1, 'atomname': 'BB', 'atype': 'P4',
                                  'resname': 'ALA', 'resid': 1,
                                  'charge_group': 1}
        assert block.nodes[1] == {'index': 2, 'atomname': 'SC1', 'atype': 'P3',
                                  'resname': 'ALA', 'resid': 1,
                                  'charge_group': 2, 'charge': -3.0}
        assert block.nodes[2] == {'index': 3, 'atomname': 'SC2', 'atype': 'P2',
                                  'resname': 'ALA', 'resid': 1,
                                  'charge_group': 3, 'charge': -3, 'mass': 72}

    @staticmethod
    def test_fixed_number_interaction():
        """
        Define an interaction for which the number of atoms required is known.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P3 1 ALA SC1 2
        3 P2 1 ALA SC2 3
        4 P2 1 ALA SC3 3

        [ bonds ]
        1 2 1 0.2 100
        2 3 4 0.6 700
        [ angles ]
        1  2  3 params
        2  3  4 params
        """
        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.gmx.itp_read.read_itp(lines, ff)
        block = ff.blocks['GLY']

        bonds = [
            vermouth.molecule.Interaction(
                atoms=[0, 1], parameters=['1', '0.2', '100'], meta={},
            ),
            vermouth.molecule.Interaction(
                atoms=[1, 2], parameters=['4', '0.6', '700'], meta={},
            ),
        ]

        angles = [
            vermouth.molecule.Interaction(
                atoms=[0, 1, 2], parameters=['params'], meta={},
            ),
            vermouth.molecule.Interaction(
                atoms=[1, 2, 3], parameters=['params'], meta={},
            ),
        ]
        assert block.interactions['bonds'] == bonds
        assert block.interactions['angles'] == angles


    @staticmethod
    def test_interaction_by_index():
        """
        Create an interaction and refer to atoms by index.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P3 1 ALA SC1 2
        3 P2 1 ALA SC2 3

        [ bonds ]
        1 2   1 0.2 100
        1 3   4 0.6 700
        """

        lines = textwrap.dedent(lines).splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.gmx.itp_read.read_itp(lines, ff)
        block = ff.blocks['GLY']

        bonds = [
            vermouth.molecule.Interaction(
                atoms=[0, 1], parameters=['1', '0.2', '100'], meta={},
            ),
            vermouth.molecule.Interaction(
                atoms=[0, 2], parameters=['4', '0.6', '700'],
                meta={},
            ),

        ]
        assert block.interactions['bonds'] == bonds

    @staticmethod
    @pytest.mark.parametrize('interaction_lines', (
        # Refers, by index, to an atom out of range
        """
        [ bonds ]
        1 6
        """,
        """
        [ bonds ]
        1 -6
        """,
        # Refers, by name, to a non-defined atom (SC2 does not exist)
        """
        [ bonds ]
        BB SC2
        """,
        # One missing atom in the definition (refers to only one atom instead of 2)
        """
        [ bonds ]
        BB
        """,
        # Prefixed atom in a block interaction
        """
        [ bonds ]
        BB +SC1
        """,
        """
        [ bonds ]
        BB -SC1
        """,
        """
        [ bonds ]
        BB <SC1
        """,
        """
        [ bonds ]
        BB >SC1
        """,
    ))
    def test_interaction_fail_reference(interaction_lines):
        """
        Faulty atom references in interactions lead to an exception.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        ; The following atoms have names that should not be valid
        2 P4 2 TRP +SC1 1
        3 P4 2 TRP -SC1 1
        4 P4 2 TRP <SC1 1
        5 P4 2 TRP >SC1 1
        """
        lines = textwrap.dedent(lines)
        lines = lines.splitlines()
        lines += textwrap.dedent(interaction_lines).splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        with pytest.raises(IOError):
            vermouth.ffinput.read_ff(lines, ff)

    @staticmethod
    @pytest.mark.parametrize('section_name', (
        'non-edges', 'patterns', 'features', '!bonds', '!angles'
    ))
    def test_invalid_section_fails(section_name):
        """
        Using the sections that are invalid in a block fails.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1

        [ {} ]
        BB SC1
        """
        lines = textwrap.dedent(lines).format(section_name)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        with pytest.raises(IOError):
            vermouth.gmx.itp_read.read_itp(lines, ff)

    @staticmethod
    def test_atomname_not_unique():
        """
        Atom names are unique in an FF block but not an itp.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA BB 1
        """
        lines = textwrap.dedent(lines)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.gmx.itp_read.read_itp(lines, ff)

    @staticmethod
    def test_index_unique():
        """
        Atom indices need to be unique.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        1 P4 1 ALA SC1 1
        """
        lines = textwrap.dedent(lines)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        with pytest.raises(IOError):
            vermouth.gmx.itp_read.read_itp(lines, ff)

    @staticmethod
    @pytest.mark.parametrize('interaction_lines, _type, expected', (
        # virtual_sitesn
        ("""
        [ virtual_sitesn ]
        ; ref func construct atoms
        6  2  1  2  3  4
        """,
        "virtual_sitesn",
        vermouth.molecule.Interaction(atoms=[5, 0, 1, 2, 3],
                                      parameters=['2'],
                                      meta={})
        ),
        # virtual_sites2
        ("""
        [ virtual_sites2 ]
        ; site  from      funct    a
        4   3  2  1  0.5
        """,
        "virtual_sites2",
        vermouth.molecule.Interaction(atoms=[3, 2, 1],
                                      parameters=['1', '0.5'],
                                      meta={})
        ),
        # virtual_sites3
        ("""
        [ virtual_sites3 ]
        4  1  2  3  1  0.810 0.810
        """,
        "virtual_sites3",
        vermouth.molecule.Interaction(atoms=[3, 0, 1, 2],
                                      parameters=['1', '0.810', '0.810'],
                                      meta={})
        ),
        # virtual_sites4
        ("""
        [ virtual_sites4 ]
        4   1  2  3  5  1  0.810 0.810 0.2
        """,
        "virtual_sites4",
        vermouth.molecule.Interaction(atoms=[3, 0, 1, 2, 4],
                                      parameters=['1', '0.810', '0.810', '0.2'],
                                      meta={})
        )
    ))
    def test_virtual_sites(interaction_lines, _type, expected):
        """
        test if index and atoms are curretly distinguished for
        this type of interaction
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1
        3 P4 1 ALA SC2 1
        4 P4 1 ALA SC3 1
        5 P4 1 ALA SC4 1
        6 VS 1 ALA SC5 1
        """
        lines = textwrap.dedent(lines)
        lines = lines.splitlines()
        lines += textwrap.dedent(interaction_lines).splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.gmx.itp_read.read_itp(lines, ff)

        assert ff.blocks['GLY'].interactions[_type][0] == expected


    @staticmethod
    @pytest.mark.parametrize('def_statements, bonds', (
        (
            """
            #ifdef FLEXIBLE
            [ bonds ]
            1  2
            2  3
            #endif
            """,
            [vermouth.molecule.Interaction(atoms=[0, 1], parameters=[], meta={"ifdef":"FLEXIBLE"}),
             vermouth.molecule.Interaction(atoms=[1, 2], parameters=[], meta={"ifdef":"FLEXIBLE"})]
        ),
        (
            """
            #ifndef FLEXIBLE
            [ bonds ]
            1   2
            2   3
            #endif
            """,
            [vermouth.molecule.Interaction(atoms=[0, 1], parameters=[], meta={"ifndef":"FLEXIBLE"}),
             vermouth.molecule.Interaction(atoms=[1, 2], parameters=[], meta={"ifndef":"FLEXIBLE"})]
        ),
        (
            """
            [ bonds ]
            1   2
            #ifdef FLEXIBLE
            2   3
            #endif
            3  4
            """,
            [vermouth.molecule.Interaction(atoms=[0, 1], parameters=[], meta={}),
             vermouth.molecule.Interaction(atoms=[1, 2], parameters=[], meta={"ifdef": "FLEXIBLE"}),
             vermouth.molecule.Interaction(atoms=[2, 3], parameters=[], meta={})]
        ),
    ))
    def test_defs(def_statements, bonds):
        """
        test the handling if ifdefs and ifndefs at
        different positions in itp-file
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1
        3 P4 1 ALA SC2 1
        4 P4 1 ALA SC3 1
        """

        new_lines = lines + def_statements
        new_lines = textwrap.dedent(new_lines)
        new_lines = new_lines.splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.gmx.itp_read.read_itp(new_lines, ff)
        assert ff.blocks['GLY'].interactions['bonds'] == bonds

    @staticmethod
    @pytest.mark.parametrize('pragma_fail_statements', (
        """
        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1
        [ bonds ]
        1   2
        #fdef FLEXIBLE
        2   3
        #endif
        3  4
        """,
        """
        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1
        3 P4 1 ALA SC2 1
        4 P4 1 ALA SC3 1
        [ bonds ]
        1   2
        #ifdef FLEXIBLE
        2   3
        3  4
        """,
        """
        [ atoms ]
        1   2
        #ifdef FLEXIBLE
        2   3
        #endif
        3  4
        """,
        """
        [ atoms ]
        1 P4 1 ALA SC1 1
        2 P4  1 ALA SC1 1
        [ bonds ]
        #include random
        """,
        """
        [atoms]
        1 P4  1 ALA SC1 1
        2 P4  1 ALA SC1 1
        [ bonds ]
        1   2   A B
        #endif
        """,
        """
        [ atoms ]
        1 P4  1 ALA SC1 1
        2 P4  1 ALA SC1 1
        3 P4  1 ALA SC1 1
        [ bonds ]
        #ifdef FLEXIBLE
        1   2   A  B
        #ifdef RANDOM
        1    3  A   B
        """,
        """
        [ atoms ]
        1 P4 1 ALA SC1 1
        2 P4  1 ALA SC1 1
        [ bonds ]
        1   2 A B
        #if
        """
    ))
    def test_pragma_fails(pragma_fail_statements):
        """
        test if incorrect use of pragmas raises
        appropiate error
        """
        lines = """
        [ moleculetype ]
        GLY 1
        """
        new_lines = lines + pragma_fail_statements
        new_lines = textwrap.dedent(new_lines)
        new_lines = new_lines.splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        with pytest.raises(IOError):
            vermouth.gmx.itp_read.read_itp(new_lines, ff)

    @staticmethod
    def test_multiple_interactions():
        """
        test if we can multiple interactions
        using the same atom indices but different
        parameters.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1
        3 P4 1 ALA SC2 1
        4 P4 1 ALA SC3 1
        [ dihedrals ]
        1   2   3   4  1  A
        1   2   3   4  1  B
        1   2   3   4  1  C
        """

        lines = textwrap.dedent(lines)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.gmx.itp_read.read_itp(lines, ff)

        dih = [vermouth.molecule.Interaction(atoms=[0, 1, 2, 3], parameters=['1', 'A'], meta={}),
               vermouth.molecule.Interaction(atoms=[0, 1, 2, 3], parameters=['1', 'B'], meta={}),
               vermouth.molecule.Interaction(atoms=[0, 1, 2, 3], parameters=['1', 'C'], meta={})]

        assert ff.blocks['GLY'].interactions['dihedrals'] == dih

    @staticmethod
    def test_exclusions():
        """
        test if can read exclusions with
        variable number of excluded atoms properly.
        """
        lines = """
        [ moleculetype ]
        GLY 1

        [ atoms ]
        1 P4 1 ALA BB 1
        2 P4 1 ALA SC1 1
        3 P4 1 ALA SC2 1
        4 P4 1 ALA SC3 1
        [ exclusions ]
        1   2   3   4
        2   3   4
        3   4
        """

        lines = textwrap.dedent(lines)
        lines = lines.splitlines()

        ff = vermouth.forcefield.ForceField(name='test_ff')
        vermouth.gmx.itp_read.read_itp(lines, ff)

        excl = [vermouth.molecule.Interaction(atoms=[0, 1, 2, 3], parameters=[], meta={}),
                vermouth.molecule.Interaction(atoms=[1, 2, 3], parameters=[], meta={}),
                vermouth.molecule.Interaction(atoms=[2, 3], parameters=[], meta={})]

        assert ff.blocks['GLY'].interactions['exclusions'] == excl


    @staticmethod
    @pytest.mark.parametrize('test_lines', (
        """[ atoms ]
        1 P4 1 ALA BB 1
        1 P4 1 ALA BB 1
        """,
        """[ atoms ]
        1 P4 1 ALA BB 1
        [ bonds ]
        10  15   A
        """,
        """[ atoms ]
        -1   P4 1 ALA BB 1
        """,
        """[ atoms ]
        AA   P4   ALA BB 1
        """
    ))
    def test_atom_block_errors(test_lines):
        """
        test if incorrectly formatted ifdefs raise
        appropriate error
        """
        lines = """
        [ moleculetype ]
        GLY 1
        """
        new_lines = lines + test_lines
        new_lines = textwrap.dedent(new_lines)
        new_lines = new_lines.splitlines()
        ff = vermouth.forcefield.ForceField(name='test_ff')
        with pytest.raises(IOError):
            vermouth.gmx.itp_read.read_itp(new_lines, ff)



# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Tests for the log helpers
"""

import logging
import string

from hypothesis import strategies as st
from hypothesis import given, note, event, example
import pytest

from vermouth.log_helpers import (get_logger, TypeAdapter, StyleAdapter,
                                  PassingLoggerAdapter, Message,
                                  BipolarFormatter, CountingHandler,
                                  ignore_warnings_and_count,)

# Pylint does *not* like pytest fixtures... Also, sometimes you just need more
# data
# pylint: disable=redefined-outer-name, too-many-arguments

KWARG_ST = st.text(alphabet=string.ascii_letters, min_size=1)


class FormatCounter(str):
    def __init__(self, *args, **kwargs):
        super().__init__()
        self._format_count = 0

    def format(self, *args, **kwargs):
        """"Increments the counter, and format self."""
        self._format_count += 1
        return self.__class__(super().format(*args, **kwargs))

    def __mod__(self, args):
        self._format_count += 1
        return self.__class__(super().__mod__(args))

    def get_count(self):
        """Returns the current value of the counter"""
        return self._format_count


class LogHandler(logging.NullHandler):
    """Helper class which will run a test for every log record"""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.test = lambda: None

    def set_test(self, test):
        """Set the test to be run"""
        self.test = test

    def handle(self, record):
        """Does nothing, except run the test."""
        self.test(record)


@pytest.fixture
def logger(request):
    """Sets up a logger at loglevel 1 and attaches a LogHandler."""
    logger_ = logging.getLogger(request.function.__name__)
    logger_.setLevel(1)
    return logger_


@pytest.fixture(scope='module')
def handler():
    """Sets up a LogHandler"""
    handler_ = LogHandler(1)
    return handler_


@given(name=st.text())
def test_get_logger(name):
    """
    Make sure get_logger gives the correct logger.
    """
    default_logger = logging.getLogger(name)
    vm_logger = get_logger(name)
    assert vm_logger.logger is default_logger


@example(args=[], type_='general', default_type='general', extra=None)
@given(
    args=st.lists(st.text(), min_size=0, max_size=5),
    type_=st.one_of(st.none(), st.text()),
    default_type=st.text(min_size=1),
    extra=st.one_of(
        st.none(),
        st.dictionaries(KWARG_ST, st.integers(), min_size=0, max_size=5)
    )
)
def test_type_adapter(logger, handler, args, type_, default_type, extra):
    """Make sure the TypeAdapter sets the correct type attr"""
    def test(record):
        """Make sure the type attribute is as expected"""
        rectype = getattr(record, 'type', None)
        if rectype is None:
            assert False, "No type for record!"
        if type_ is None:
            assert rectype == default_type
        else:
            assert rectype == type_
        if extra is not None:
            note(str(dir(record)))
            for key, val in extra.items():
                assert getattr(record, key, None) == val

    logger.addHandler(handler)
    logger = TypeAdapter(logger, default_type=default_type, extra=extra)
    handler.set_test(test)
    fmt = ['%s'] * len(args)
    fmt = ' '.join(fmt)

    note('fmt={}'.format(fmt))

    if type_ is None:
        logger.info(fmt, *args)
        event('type is None')
    else:
        logger.info(fmt, *args, type=type_)


@example(args=[], kwargs={}, type_='general', default_type='general', extra=None)
@given(
    args=st.lists(st.text(), min_size=0, max_size=5),
    kwargs=st.dictionaries(KWARG_ST,
                           st.text(), min_size=0, max_size=5),
    type_=st.one_of(st.none(), st.text()),
    default_type=st.text(min_size=1),
    extra=st.one_of(
        st.none(),
        st.dictionaries(KWARG_ST, st.integers(), min_size=0, max_size=5)
    )
)
def test_style_type_adapter(logger, handler, args, kwargs, type_, default_type, extra):
    """Make sure that if you have both a TypeAdapter and a StyleAdapter the
    type you provide ends up in the right place, and that it doesn't interfere
    with keyword formatting."""
    def test(record):
        """Make sure the type attribute is as expected"""
        rectype = getattr(record, 'type', None)
        if rectype is None:
            assert False, "No type for record!"
        if type_ is None:
            assert rectype == default_type
        else:
            assert rectype == type_
        if extra is not None:
            for key, val in extra.items():
                assert getattr(record, key, None) == val

    logger.addHandler(handler)
    logger = TypeAdapter(logger, default_type=default_type)
    logger = StyleAdapter(logger, extra=extra)
    handler.set_test(test)
    fmt = ['{}']*len(args) + ['{' + name + '}' for name in kwargs]
    fmt = ' '.join(fmt)
    print('-' * 60)

    note('fmt={}'.format(fmt))
    if type_ is None:
        logger.info(fmt, *args, **kwargs)
        event('type is None')
    else:
        logger.info(fmt, *args, type=type_, **kwargs)


@example(args=[], kwargs={}, extra=None)
@given(
    args=st.lists(st.text(), min_size=0, max_size=5),
    kwargs=st.dictionaries(KWARG_ST,
                           st.text(), min_size=0, max_size=5),
    extra=st.one_of(st.none(),
                    st.dictionaries(KWARG_ST, st.integers(), min_size=0, max_size=5))
)
def test_style_adapter(logger, handler, args, kwargs, extra):
    """Make sure the StyleAdapter can do keyword formatting"""
    def test(record):
        """Make sure the formatting worked"""
        assert record.getMessage() == expected
        if extra is not None:
            for key, val in extra.items():
                assert getattr(record, key, None) == val

    logger.addHandler(handler)
    logger = StyleAdapter(logger, extra=extra)
    handler.set_test(test)
    fmt = ['{}']*len(args) + ['{' + name + '}' for name in kwargs]
    fmt = ' '.join(fmt)

    note('fmt={}'.format(fmt))
    expected = fmt.format(*args, **kwargs)
    logger.info(fmt, *args, **kwargs)


@given(
    args=st.lists(st.text(), min_size=0, max_size=5),
    kwargs=st.dictionaries(KWARG_ST, st.text(), min_size=1, max_size=5),
)
def test_passing_adapter(logger, handler, args, kwargs):
    """Make sure the PassingLoggerAdapter does not allow keywords to be set for
    formatting."""
    def test(_):
        assert False
    handler.set_test(test)
    logger.addHandler(handler)
    logger = PassingLoggerAdapter(logger)
    fmt = ['%s']*len(args) + ['%(' + name + ')s' for name in kwargs]
    fmt = ' '.join(fmt)
    note(fmt)
    logger.setLevel(logging.INFO)
    logger.debug('')


@given(
    args=st.lists(st.text(), min_size=0, max_size=5),
    kwargs=st.dictionaries(KWARG_ST, st.text(), min_size=0, max_size=5)
)
def test_message(args, kwargs):
    """Make sure Message doesn't formats it's contents needlessly"""
    fmt = ['{}']*len(args) + ['{' + name + '}' for name in kwargs]
    fmt = ' '.join(fmt)
    note('fmt={}'.format(fmt))
    counter = FormatCounter(fmt)
    msg = Message(counter, args, kwargs)
    assert counter.get_count() == 0
    assert str(msg) == fmt.format(*args, **kwargs)
    assert counter.get_count() == 1

    assert repr(msg) == '"{}".format(*{}, **{})'.format(fmt, args, kwargs)


def test_bipolar_formatter():
    """
    Make sure the bipolar formatter calls the correct formatter.
    """
    low_counter = FormatCounter('%(message)s')
    low_formatter = logging.Formatter(fmt=low_counter)
    high_counter = FormatCounter('%(message)s')
    high_formatter = logging.Formatter(fmt=high_counter)
    logger = logging.getLogger('test_bipolar_formatter')
    formatter = BipolarFormatter(low_formatter, high_formatter,
                                 logging.INFO, logger=logger)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger.addHandler(handler)

    assert logger is formatter.logger

    logger.setLevel(1)

    logger.info('boo')
    logger.error('baa')
    # Logger level <= cutoff, so both should go to low_counter
    assert low_counter.get_count() == 2
    assert high_counter.get_count() == 0

    logger.setLevel(logging.WARNING)
    logger.info('boo')
    logger.warning('baa')
    # Logger level > cutoff, so both should go to high_counter, but one of the
    # messages is too low priority to be shown.
    assert low_counter.get_count() == 2
    assert high_counter.get_count() == 1


def test_bipolar_formatter_logger():
    """
    Make sure the bipolar logger picks the correct logger if none is given.
    """
    low_counter = FormatCounter('%(message)s')
    low_formatter = logging.Formatter(fmt=low_counter)
    high_counter = FormatCounter('%(message)s')
    high_formatter = logging.Formatter(fmt=high_counter)
    formatter = BipolarFormatter(low_formatter, high_formatter,
                                 logging.INFO)
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger = formatter.logger
    logger.addHandler(handler)

    expected_logger = logging.getLogger()
    assert expected_logger is logger


@pytest.mark.parametrize('level, type_, expected', (
    [None, None, 7],
    [logging.DEBUG, None, 2],
    [None, 'a', 2],
    [logging.INFO, 'general', 2],
    [None, 'd', 0]
))
def test_counter(level, type_, expected):
    """Make sure the CountingHandler has learned how to count"""
    logger = logging.getLogger('test_counter')
    handler = CountingHandler(default_type='0')
    logger.addHandler(handler)
    logger.setLevel(1)
    logger = TypeAdapter(logger)

    logger.info('')
    logger.info('')
    logger.info('', type='c')
    logger.warning('', type='a')
    logger.error('', type='a')
    logger.debug('', type='b')
    logger.debug('', type='b')
    expected_total = {logging.DEBUG: {'b': 2},
                      logging.INFO: {'general': 2, 'c': 1},
                      logging.WARNING: {'a': 1},
                      logging.ERROR: {'a': 1}}
    assert handler.counts == expected_total

    assert handler.number_of_counts_by(level=level, type=type_) == expected


@pytest.fixture
def mock_counter():
    class MockCountingHandler:
        def __init__(self):
            self.counts = {
                logging.WARNING: {
                    'default': 4,
                    'something': 8,
                    'other thing': 2,
                    'empty': 0,
                },
                logging.INFO: {
                    'something': 5,
                },
            }

        def number_of_counts_by(self, level=None, type=None):
            assert type is None, 'The mock does not discriminate by type'
            results = {
                logging.WARNING: 14,
                logging.INFO: 5,
            }
            return results.get(level, 0)

    return MockCountingHandler()


@pytest.mark.parametrize('specification, level, expected', (
    # -maxwarn 4
    ([[(None, 4)]], logging.WARNING, 10),
    # -maxwarn something
    ([[('something', None)]], logging.WARNING, 6),
    # -maxwarn something:5
    ([[('something', 5)]], logging.WARNING, 9),
    # -maxwarn something:5 2
    ([[('something', 5), (None, 2)]], logging.WARNING, 7),
    # -maxprint something:5 -maxprint 2
    ([[('something', 5)], [(None, 2)]], logging.WARNING, 7),
    # Not accessible from the command line
    ([[('something', 3)]], logging.INFO, 2),
))
def test_ignore_warnings_and_count(mock_counter, specification, level, expected):
    remaining = ignore_warnings_and_count(mock_counter, specification, level)
    assert remaining == expected





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Copyright 2018 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tests for vermouth.forcefield.get_native_force_field.

These tests cannot be in the same module as the rest of the tests for the
vermouth.forcefield module. Indeed, the tests in this file causes to load all
force fields which create test coverage noise.
"""

import pytest
import vermouth.forcefield


def test_get_native_force_field_identity():
    """
    Test that the same instance is returned for multiple request of the same
    force field.
    """
    # We could parametrize the test to check each force field separatly, but we
    # need to make sure that different requested name result in different
    # instances. Without that, the test would pass if the function was always
    # returning the same object regardless of the argument.
    requests = ('universal', 'martini22', 'martini22p')
    first_call = set(id(vermouth.forcefield.get_native_force_field(name))
                     for name in requests)
    second_call = set(id(vermouth.forcefield.get_native_force_field(name))
                      for name in requests)

    # Are all the force fields different?
    assert len(first_call) == len(second_call) == len(requests)
    # Are the same instances returned for the two calls?
    assert first_call == second_call


def test_get_native_force_field_key_error():
    """
    Test that get_native_force_field raises a KeyError for unfound force fields.
    """
    with pytest.raises(KeyError):
        vermouth.forcefield.get_native_force_field('non existant')



# Copyright 2020 University of Groningen
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Provides the DeferredFileWriter, which allow writing of files without affecting
existing files, until it is clear the written changes are correct.
"""


from builtins import open as _open

import collections
import os
import pathlib
import shutil
import threading
import tempfile

from .log_helpers import StyleAdapter, get_logger

LOGGER = StyleAdapter(get_logger(__name__))

lock = threading.Lock()


class Singleton(type):
    """
    Metaclass for creating singleton objects. Taken from [1]_.

    .. [1] https://stackoverflow.com/questions/50566934/why-is-this-singleton-implementation-not-thread-safe/50567397
    """
    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            with lock:
                if cls not in cls._instances:
                    cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


class DeferredFileWriter(metaclass=Singleton):
    """
    A singleton class/object that is intended to prevent writing output to files
    that is invalid, due to e.g. warnings further down the pipeline.

    If this class is used to open a file for writing, a temporary file is created
    and returned instead. Once it's clear the output produced is valid the
    :meth:`~vermouth.file_writer.DeferredFileWriter.write` method can be used to
    finalize the written changes by moving them to their intended destination.
    If a file with that name already exists it is backed up according to the
    Gromacs scheme.
    """
    def __init__(self):
        self.open_files = collections.deque()
        self._tmpdir = None

    def open(self, filename, mode='r', *args, **kwargs):
        """
        If mode is either 'w' or 'a', opens and returns a handle to a temporary
        file. If mode is 'r' opens and returns a handle to the file specified.

        Once :meth:`~vermouth.file_writer.DeferredFileWriter.write` is called
        the changes written to all files opened this way are propagated to their
        final destination.

        Parameters
        ----------
        filename: os.PathLike
            The final name of the file to be opened.
        mode: str
            The mode in which the file is to be opened.
        *args: collections.abc.Iterable
            Passed to :func:`os.fdopen`.
        **kwargs: dict
            Passed to :func:`os.fdopen`.

        Returns
        -------
        io.IOBase
            An opened file
        """
        path = pathlib.Path(filename)
        # Make the path absolute, in case the current working directory is
        # changed between now and writing. Can't do path.resolve() due to py35
        # requiring the file to exist.
        path = path.parent.resolve() / path.name
        # Let's see if we already opened this file. If so, get the corresponding
        # temporary file.
        for tmp_path, open_path, _ in self.open_files:
            # Can't use Path.samefile, since the files don't have to exist yet
            if open_path == path:
                return _open(tmp_path, mode, *args, **kwargs)

        if '+' in mode or 'a' in mode or 'w' in mode:  # Append and write
            return self._open_tmp_file(path, *args, mode=mode, **kwargs)
        elif 'r' in mode:  # Read, do nothing special
            return _open(filename, mode, *args, **kwargs)
        raise KeyError('Unknown file mode.')

    def _open_tmp_file(self, filename, mode='w', *args, **kwargs):
        suffix = filename.suffix
        with lock:
            handle, tmp_path = tempfile.mkstemp(suffix=suffix, dir=self._tmpdir)
        self.open_files.append([tmp_path, filename, mode])
        if '+' in mode and 'r' in mode:
            # If r+, preserve original file contents. Otherwise, truncate.
            shutil.copy2(str(filename), tmp_path)
        return os.fdopen(handle, mode, *args, **kwargs)

    @staticmethod
    def _find_free_path(file_path):
        """
        Find the first free (backup) path that looks like `file_path`. If
        file_path does not exist, returns file_path. Else, generates
        `#{name}.{idx}#`, incrementing idx until no file exists.

        Parameters
        ----------
        file_path: pathlib.PathLike

        Returns
        -------
        pathlib.Path
            The first path that does not exist yet
        """
        file_path = pathlib.Path(file_path)
        backup_path = pathlib.Path(file_path)
        idx = 1
        while backup_path.exists():
            backup_path = file_path.with_name('#{name}.{idx}#'.format(name=file_path.name, idx=idx))
            idx += 1
        return backup_path

    def write(self):
        """
        Finalize writing all open files by moving the created temporary files to
        their final destinations.

        Existing file destinations will be backed up according to the Gromacs
        scheme.
        """
        while self.open_files:
            tmp_path, final_path, mode = self.open_files.popleft()
            if 'w' in mode or '+' in mode:  # write
                self._write_file(tmp_path, final_path)
            elif 'r' in mode:  # read = error
                raise AssertionError("Files opened with mode 'r' should not be "
                                     "treated special")
            elif 'a' in mode:  # append
                self._append_file(tmp_path, final_path, mode)
            else:
                raise KeyError('Unknown file mode')

    def _write_file(self, tmp_path, final_path):
        # There is no way to move a file and make it error if the destination
        # already exists, so use a lock instead.
        with lock:
            free_path = self._find_free_path(final_path)
            if free_path != final_path:
                LOGGER.info('Backing up {} to {}.', final_path, free_path, type='general')
                shutil.move(str(final_path), str(free_path))
            LOGGER.debug('Writing output to {}.', final_path, type='general')
            shutil.move(tmp_path, str(final_path))

    @staticmethod
    def _append_file(tmp_path, final_path, mode='a'):
        """
        Append the contents of tmp_path to final_path and remove tmp_path.
        """
        if 'b' in mode:
            tmp_mode = 'rb'
        else:
            tmp_mode = 'r'
        with _open(str(final_path), mode=mode) as final_file, _open(tmp_path, mode=tmp_mode) as tmp_file:
            final_file.write(tmp_file.read())
        os.remove(tmp_path)

    def close(self):
        """
        Remove all produced temporary files.
        """
        while self.open_files:
            tmp_path, *_ = self.open_files.popleft()
            try:
                os.remove(tmp_path)
            except FileNotFoundError:
                pass

    def __del__(self):
        self.close()
        # super().__del__()  # object has no __del__


open = DeferredFileWriter().open
